import logger from 'logger/logger'
import { AppleInApp, ApplePayStatus } from 'modules/AppleInApp'
import { InAppRecord } from 'modules/InAppRecord'
import { IPayResult, reportApplePurchaseResult } from 'service/game.svr'
import { IosPaySvr } from 'service/iospay.svr'
import { BaseController, router, ZError, role, ROLE_ANON } from 'zutils'

class ApplePayController extends BaseController {
  @router('post /pay/apple/verify')
  @router('post /native_pay/apple/verify')
  async verifyApplePay(req, res) {
    logger.db('applepay_verify', req)
    const { openId } = req.user
    let { list } = req.params
    if (typeof list === 'string') {
      try {
        list = JSON.parse(list)
      } catch (err) {
        throw new ZError(10, 'purchase data is empty')
      }
    }
    if (!list || !list.length) {
      throw new ZError(10, 'purchase data is empty')
    }
    logger.info(`verify ios purchase::list=${list}`)
    let results: IPayResult[] = []
    let ids: Set<string> = new Set()
    for (let sub of list) {
      let status = ApplePayStatus.PENDING
      let tx = await new IosPaySvr().getTransactionInfo(sub)
      if (!tx) {
        continue
      }
      await InAppRecord.addRecord(openId, tx, 'apple')
      logger.info('getTransactionInfo:: ', tx)
      if (tx.transactionReason === 'PURCHASE' && tx.inAppOwnershipType === 'PURCHASED') {
        status = ApplePayStatus.SUCCESS
      }
      if (status === ApplePayStatus.SUCCESS) {
        let record = await AppleInApp.insertOrUpdate(
          { outOrderId: tx.transactionId },
          {
            account: openId,
            productId: tx.productId,
            gameOrderId: tx.appAccountToken,
            country: tx.storefront,
            status,
            kind: tx.type,
            purchaseTime: tx.purchaseDate,
            data: tx,
            amount: tx.quantity,
            appid: tx.bundleId,
            env: tx.environment,
            $inc: { version: 1 },
          },
        )
        ids.add(sub)
        if (!record.reportStatus) {
          results.push({
            productId: record.productId,
            gameOrderId: record.gameOrderId,
            orderId: record.outOrderId,
            status: record.status,
          })
        }
      }
    }
    if (results.length) {
      logger.info(`need report apple purchase result::results=${JSON.stringify(results)}`)
      setImmediate(async () => {
        try {
          await reportApplePurchaseResult(results)
        } catch (err) {
          logger.error('report apple purchases failed', err.message || err)
        }
      })
    }
    list.forEach((sub: string) => ids.add(sub))
    return [...ids]
  }

  @role(ROLE_ANON)
  @router('get /pay/apple/info/:tid')
  async applePayInfo(req, res) {
    const { tid } = req.params
    if (!tid) {
      throw new ZError(10, 'purchase data is empty')
    }
    let transaction = await new IosPaySvr().getTransactionInfo(tid)
    logger.info('getTransactionInfo:: ', transaction)
    return transaction
  }

  /**
   * apple in-app purchase server-to-server notification
   * use notification v2
   * https://developer.apple.com/documentation/appstoreservernotifications/app_store_server_notifications_v2
   */
  @role(ROLE_ANON)
  @router('post /pay/out/apple/notify')
  async notifyApplePay(req, res) {
    let { signedPayload } = req.params
    await new IosPaySvr().verifyNotification(signedPayload)
    return res.send('ok')
  }
}
