import axios from 'axios'
import { hmacSha256, md5, sha1 } from 'zutils/utils/security.util'
import crypto from 'crypto'
import { generateKVStr } from 'zutils/utils/net.util'
import logger from 'logger/logger'
import { GasSvr } from './gas.svr'
import { toBigInt, toBigWei } from 'zutils/utils/bn.util'
import { PriceSvr } from './price.svr'

export function createSimpleSign(data: any) {
  let timestamp = Date.now()
  let appid = process.env.ALCHEMY_APPID
  let secret = process.env.ALCHEMY_APP_SECRET
  let signData = { appid, timestamp }
  signData = Object.assign(signData, data)
  let signStr = Object.keys(signData)
    .sort()
    .map(key => `${key}=${signData[key]}`)
    .join('&')
  let sign = hmacSha256(signStr, secret).toLowerCase()
  return {
    appid,
    timestamp,
    sign,
  }
}

export function createSha1Sign() {
  let timestamp = Date.now()
  let appid = process.env.ALCHEMY_APPID
  let secret = process.env.ALCHEMY_APP_SECRET
  let sign = sha1(appid + secret + timestamp)
  return {
    appid,
    timestamp,
    sign,
  }
}
const signParamArr = [
  'network',
  'crypto',
  'address',
  'fiat',
  'fiatAmount',
  'payWayCode',
  'country',
  'accountId',
  'orderId',
  'timestamp',
  'salt',
]
/**
 * 游戏服签名验证
 * @param params
 * @returns
 */
export function checkPrePaySign(params: any): boolean {
  let signArr = []
  signParamArr.forEach(key => {
    signArr.push(params[key])
  })
  const signToCheck = md5(signArr.join('&') + process.env.GAME_PAY_KEY)
  return params.sign === signToCheck
}

/**
 * Check if the pay result sign is valid
 * @param data - the data to be checked
 * @returns true if the sign is valid, false otherwise
 */
export function checkPayResultSign(data: any) {
  const { appId, appid, orderNo, crypto, network, address, signature } = data
  const appIdReal = appId || appid
  const sign = sha1(appIdReal + process.env.ALCHEMY_APP_SECRET + appIdReal + orderNo + crypto + network + address)
  return sign === signature
}

export function checkSimpleSign(headers: any, data: any) {
  // alchemy 很不严谨, 有时候是 appid, 有时候是 appId
  const { appid, appId, timestamp, sign } = headers
  let appIdToCheck = appId || appid
  let signData = { appid: appIdToCheck, timestamp }
  signData = Object.assign(signData, data)
  let signStr = Object.keys(signData)
    .sort()
    .map(key => `${key}=${signData[key]}`)
    .join('&')
  const expectedSign = hmacSha256(signStr, process.env.ALCHEMY_APP_SECRET).toLowerCase()
  logger.info('compare sign: ', sign, expectedSign)
  // const expectedSign = sha1(appIdToCheck + process.env.ALCHEMY_APP_SECRET + timestamp)
  return sign === expectedSign
}

/**
 * Create page sign
 * @param plainText - plain text to be encrypted
 * @returns encrypted text
 */
export function createPageSign(plainText: string) {
  let secret = process.env.ALCHEMY_APP_SECRET
  try {
    const plainTextData = Buffer.from(plainText, 'utf8')
    const secretKey = Buffer.from(secret, 'utf8')
    const iv = secret.substring(0, 16)

    const cipher = crypto.createCipheriv('aes-128-cbc', secretKey, iv)

    let encrypted = cipher.update(plainTextData)
    encrypted = Buffer.concat([encrypted, cipher.final()])

    return encrypted.toString('base64')
  } catch (e) {
    console.log(`AES encrypting exception, msg is ${e.toString()}`)
  }
  return null
}

export function checkSha1Sign(headers: any) {
  const { appid, appId, timestamp, sign } = headers
  // timestamp 5分钟内有效
  if (parseInt(timestamp) < Date.now() - 1000 * 60 * 5) {
    return false
  }
  let appIdToCheck = appId || appid || process.env.ALCHEMY_APPID
  const expectedSign = sha1(appIdToCheck + process.env.ALCHEMY_APP_SECRET + timestamp)
  return sign === expectedSign
}
/**
 * Refresh token
 * https://alchemycn.readme.io/docs/获取token
 * @param email - user email
 * @returns token
 */
export async function refreshToken(email: string) {
  const data = JSON.stringify({ email })
  const { appid, timestamp, sign } = createSimpleSign({ email })
  const host = process.env.ALCHEMY_API_BASE
  const config = {
    method: 'post',
    url: `${host}/open/api/v3/merchant/getToken`,
    headers: {
      appId: appid,
      timestamp: timestamp,
      sign: sign,
      'Content-Type': 'application/json',
    },
    data: data,
  }
  let response = await axios(config)
  return response.data
}
/**
 * 创建订单
 * https://alchemycn.readme.io/docs/创建订单
 */
export async function createOrder(token: string, data: any) {
  const { appid, timestamp, sign } = createSimpleSign(data)
  const host = process.env.ALCHEMY_API_BASE
  const config = {
    method: 'post',
    url: `${host}/open/api/v3/merchant/trade/create`,
    headers: {
      'access-token': token,
      appId: appid,
      timestamp: timestamp,
      sign: sign,
      'Content-Type': 'application/json',
    },
    data: data,
  }
  let response = await axios(config)
  return response.data
}

/**
 * 更新订单状态
 * 当您给用户转移token后，您需要通过此接口通知Alchemy打币的信息
 * https://alchemycn.readme.io/docs/更新订单状态
 * TODO:: test
 */
export async function updateOrderStatus(data: any) {
  const { appid, timestamp, sign } = createSha1Sign()
  const url = process.env.ALCHEMY_API_BASE + '/webhooks/treasure'
  const config = {
    method: 'post',
    url,
    headers: {
      'Content-Type': 'application/json',
      appId: appid,
      sign: sign,
      timestamp,
    },
    data,
  }
  let response = await axios(config)
  return response.data
}
/**
 * 查询数字货币币价
 * https://alchemycn.readme.io/docs/查询数字货币币价
 */
export async function queryPrice(data: any) {
  let dataOrign = {
    crypto: 'ETH',
    network: 'ETH',
    fiat: 'USD',
    amount: '100',
    payWayCode: '10001',
    side: 'BUY',
  }
  dataOrign = { ...dataOrign, ...data }
  const { appid, timestamp, sign } = createSimpleSign(dataOrign)
  const config = {
    method: 'post',
    url: process.env.ALCHEMY_API_BASE + '/open/api/v3/merchant/order/quote',
    headers: {
      'Content-Type': 'application/json',
      appId: appid,
      sign,
      timestamp,
    },
    data: dataOrign,
  }
  let response = await axios(config)
  return response.data
}

/**
 * 查询法币及支付方式
 * https://alchemycn.readme.io/docs/查询法币及支付方式
 * @returns
 */
export async function queryFiat() {
  let dataOrign = {
    type: 'BUY',
  }
  const { appid, timestamp, sign } = createSimpleSign(dataOrign)
  let url = process.env.ALCHEMY_API_BASE + '/open/api/v3/merchant/fiat/list'
  url = generateKVStr({ data: dataOrign, encode: true, uri: url })
  const config = {
    method: 'GET',
    url,
    headers: {
      'Content-Type': 'application/json',
      appId: appid,
      sign,
      timestamp,
    },
  }
  let response = await axios(config)
  return response.data
}
/**
 * 计算发币需要的gas费, 换算成对应的数字货币
 * 返回的数值需要fromGwei转换成wei
 * @param cryptoPrice
 * @returns
 */
export async function calcNetworkFee(crypto: string, cryptoPrice: string, ethPrice?: string) {
  let tmp = crypto.toLowerCase() === process.env.CF_ETH_NAME.toLowerCase() ? 'ETH' : crypto
  let gasLimit = await new GasSvr().estimateGas({ crypto: tmp, network: 'ARBITRUM' })
  // let gas = gas * 0.0000000001
  const gasPriceArb = '100,000,000' // arbitrum one 上gasprice固定为0.1Gwei
  let eth = toBigInt(gasLimit) * toBigInt(gasPriceArb)
  // 20230718 沟通, cryptoAmount包含了gas费, 需要减去后再发送
  if (!ethPrice) {
    ethPrice = await new PriceSvr().fetchPrice({ crypto: 'ETH', network: 'ARBITRUM', fiat: 'USD' })
  }
  let networkFee = (eth * toBigWei(ethPrice)) / toBigWei(cryptoPrice)
  return { networkFee, ethPrice }
}
