import { shortUuid } from 'zutils/utils/security.util'
import { ZError, singleton } from 'zutils'

export interface IQrData {
  clientId: string
  qrId: string
  expire: number
  remove: number
  status: number
  userId?: string
  key?: string
  token?: string
}

const DEFAULT_EXPIRE_TIME = 5 * 60 * 1000
const DEFAULT_DELETE_TIME = 10 * 60 * 1000

@singleton
export class BridgeSvr {
  private clientMap: Map<string, IQrData> = new Map()
  private qrMap: Map<string, IQrData> = new Map()

  public registClientReq(clientId: string): string {
    const qrId = shortUuid()
    const now = Date.now()
    if (this.clientMap.has(clientId)) {
      const data = this.clientMap.get(clientId)
      if (data.expire > now && data.status == 0) {
        data.expire = now + DEFAULT_EXPIRE_TIME
        data.remove = now + DEFAULT_DELETE_TIME
        return data.qrId
      } else {
        if (this.qrMap.has(data.qrId)) {
          this.qrMap.delete(data.qrId)
        }
        this.clientMap.delete(clientId)
      }
    }
    let data = {
      clientId,
      qrId,
      expire: now + DEFAULT_EXPIRE_TIME,
      remove: now + DEFAULT_DELETE_TIME,
      status: 0,
    }
    this.clientMap.set(clientId, data)
    this.qrMap.set(qrId, data)
    let self = this
    setImmediate(function () {
      self.cacheGc()
    })
    return qrId
  }

  public cacheGc() {
    let now = Date.now()
    for (let [key, data] of this.qrMap.entries()) {
      if (data.remove <= now) {
        this.qrMap.delete(key)
      }
    }
    for (let [key, data] of this.clientMap.entries()) {
      if (data.remove <= now) {
        this.clientMap.delete(key)
      }
    }
  }

  public fetchData(qrId: string) {
    const data = this.qrMap.get(qrId)
    return data
  }

  public updateData(qrId: string, key: string, token: string) {
    if (!this.qrMap.has(qrId)) {
      throw new ZError(21, 'login data not found')
    }
    let data = this.qrMap.get(qrId)
    const now = Date.now()
    if (data.status === 0 && data.expire < now) {
      throw new ZError(22, 'login data expired')
    }
    if (data.status != 0) {
      throw new ZError(23, 'already logined')
    }
    data.status = 1
    data.key = key
    data.token = token
    return true
  }
}
