import axios from 'axios'
import { PayRecordClass } from 'modules/PayRecord'
import { DocumentType } from '@typegoose/typegoose'
import { hmacSha256 } from 'zutils/utils/security.util'
import { NetClient } from 'net/NetClient'
import logger from 'logger/logger'
import { ReportQueue } from 'modules/ReportQueue'
import { AppleInApp } from 'modules/AppleInApp'
import { GoogleInApp } from 'modules/GoogleInApp'

export interface IPayResult {
  // 渠道返回的productid
  productId: string
  // 游戏服订单id
  gameOrderId: string
  // 渠道返回的订单id
  orderId: string
  status: number
}

export function assembleGameData(data: DocumentType<PayRecordClass>, mini = true) {
  if (mini) {
    return {
      account_id: data.gameAccountId,
      order_id: data.gameOrderId,
      status: data.status,
      id: data.id,
      txhash: data.txHash || '',
    }
  }
  return {
    account_id: data.gameAccountId,
    order_id: data.gameOrderId,
    status: data.status,
    id: data.id,
    txhash: data.txHash || '',
    fiat: data.fiat,
    fiat_amount: data.fiatAmount,
    country: data.country,
    payway_code: data.payWayCode || '10001',
    crypto: data.crypto,
    network: data.network,
    crypto_amount: data.cryptoSend || data.cryptoAmountEstimate,
    address: data.address,
  }
}

export function checkGameSign({ id, timestamp, sign }: { id: string; timestamp: string; sign: string }) {
  const signStr = `id=${id}&timestamp=${timestamp}`
  const hash = hmacSha256(signStr, process.env.HASH_SALT).toLowerCase()
  return sign === hash
}

export async function reportPayResult(data: DocumentType<PayRecordClass>) {
  const repData = assembleGameData(data)
  const signStr = Object.keys(repData)
    .sort()
    .map(key => `${key}=${encodeURIComponent(repData[key])}`)
    .join('&')
  const sign = hmacSha256(signStr, process.env.HASH_SALT).toLowerCase()
  const url = `${process.env.GAME_PAY_CB_URL}?c=Shop&a=buyGoodsDirect&${signStr}&sign=${sign}`
  const reqConfig: any = {
    method: 'get',
    url,
    headers: {
      'Content-Type': 'application/json',
    },
  }
  return axios(reqConfig)
}
// 上报google支付结果
export async function reportGooglePurchaseResult(records: IPayResult[]) {
  return reportPurchaseResult(records, 'google')
}

export async function reportApplePurchaseResult(records: IPayResult[]) {
  return reportPurchaseResult(records, 'apple')
}

export async function reportPurchaseResult(records: IPayResult[], channel: string) {
  const url = `${process.env.GAME_PAY_CB_URL}?c=Callback&a=dispatch&action=inappPurchase`
  let reportData: any = {
    channel,
    records,
  }
  let signStr = `channel=${channel}&`
  signStr += records
    .map(record =>
      Object.keys(record)
        .sort()
        .map(key => `${key}=${record[key]}`)
        .join('&'),
    )
    .join('&')
  const sign = hmacSha256(signStr, process.env.HASH_SALT).toLowerCase()
  reportData.sign = sign
  const reqData = {
    url,
    data: JSON.stringify(reportData),
  }
  let hasError = false
  let errMsg = ''
  try {
    let result = await new NetClient().httpPost(reqData)
    if (result.errcode) {
      hasError = true
      errMsg = `${result.errcode}: ${result.errmsg}`
    }
  } catch (err) {
    hasError = true
    errMsg = err.message || err
  }
  if (hasError) {
    for (let record of records) {
      await ReportQueue.insertOrUpdate(
        { channel: channel, recordId: record.orderId, status: record.status },
        {
          dataId: record.productId,
          orderId: record.gameOrderId,
          $push: { reps: { err: errMsg, time: Date.now() } },
        },
      )
    }
  } else {
    for (let record of records) {
      await ReportQueue.updateOne(
        { channel: channel, recordId: record.orderId, status: record.status },
        { reportStatus: 1 },
      )
      if (channel === 'apple') {
        await AppleInApp.updateOne({ outOrderId: record.orderId }, { reportStatus: 1 })
      } else if (channel === 'google') {
        await GoogleInApp.updateOne({ outOrderId: record.orderId }, { reportStatus: 1 })
      }
    }
  }
}
